@@ -4,11 +4,12 @@
  * You may distribute under the terms of either the GNU General Public
  * License or the Perl Artistic License.
  */
+#include "config.h"
+
 #include <stdio.h>
 #include <string.h>
 #include <time.h>
 
-#include "config.h"
 
 #ifndef VMS
 # include <sys/time.h>
@@ -18,23 +19,46 @@
 # include <sys/types.h>
 #endif
 
+#ifdef HAVE_TERMIOS_H
+# include <termios.h>
+#endif
+
+#ifdef VMS
+# include <unixlib.h>
+# include <unixio.h>
+# include <dvidef.h>
+# include <descrip.h>
+#else
+# if !defined(sun)
+#  include <sys/ioctl.h>
+# endif
+#endif
+
+
+#ifdef SYSV
+# include <sys/termio.h>
+# include <sys/stream.h>
+# include <sys/ptem.h>
+# include <sys/tty.h>
+#endif
+
 #include <errno.h>
 #include "slang.h"
 #include "_slang.h"
 
-#ifndef NO_STDLIB_H
+#ifdef HAVE_STDLIB_H
 # include <stdlib.h>
 #endif
 
-#ifndef NO_UNISTD_H
+#ifdef HAVE_UNISTD_H
 # include <unistd.h>
 #endif
 
-#ifdef __DECC
+#if defined(__DECC) && defined(VMS)
 /* These get prototypes for write an sleep */
 # include <unixio.h>
-# include <signal.h>
 #endif
+#include <signal.h>
 
 
 /* Colors:  These definitions are used for the display.  However, the 
@@ -68,9 +92,21 @@ int SLtt_Use_Blink_For_ACS = 0;
 int SLtt_Newline_Ok = 0;
 int SLtt_Has_Alt_Charset = 0;
 
+/* -1 means unknown */
+int SLtt_Has_Status_Line = -1;	       /* hs */
+
 static int Automatic_Margins;
 /* static int No_Move_In_Standout; */
 static int Worthless_Highlight;
+#define HP_GLITCH_CODE
+#ifdef HP_GLITCH_CODE
+/* This glitch is exclusive to HP term.  Basically it means that to clear 
+ * attributes, one has to erase to the end of the line.
+ */
+static int Has_HP_Glitch;
+#endif
+
+static char *Reset_Color_String;
 
 static int Linux_Console;
 
@@ -122,7 +158,12 @@ static Ansi_Color_Type Ansi_Color_Map[JMAX_COLORS] =
 /* work around for sco console bug that can't handle combined sequences */
   static char *Color_Escape_Sequence = "\033[3%dm\033[4%dm";
 #else
-  static char *Color_Escape_Sequence = "\033[3%d;4%dm";
+/* Believe it or not, this is what is in the linux terminfo database.  It
+ * produces the same escape sequence but it is much more CPU intensive.  
+ * Why not just encode it as "\033[3%p1%dm\033[4%p2%dm" ???
+ */
+ /* static char *Color_Escape_Sequence = "\033[%p1%{30}%+%dm\033[%p2%{40}%+%dm"; */
+static char *Color_Escape_Sequence = "\033[3%d;4%dm";
 #endif
 
 char *SLtt_Graphics_Char_Pairs;	       /* ac termcap string -- def is vt100 */
@@ -147,7 +188,7 @@ static char *Add_N_Lines_Str; /* = "\033[%dL"; */  /* AL */
 static char *Rev_Scroll_Str;
 static char *Curs_Up_Str;
 static char *Curs_F_Str;    /* RI termcap string */
-static char *Cursor_Visible_Str;    /* vs termcap string */
+static char *Cursor_Visible_Str;    /* ve termcap string */
 static char *Cursor_Invisible_Str;    /* vi termcap string */
 
 static char *Start_Alt_Chars_Str;  /* as */
@@ -155,7 +196,16 @@ static char *End_Alt_Chars_Str;   /* ae */
 static char *Enable_Alt_Char_Set;  /* eA */
 
 static char *Term_Init_Str;
+static char *Keypad_Init_Str;
 static char *Term_Reset_Str;
+static char *Keypad_Reset_Str;
+
+/* status line functions */
+static char *Disable_Status_line_Str;  /* ds */
+static char *Return_From_Status_Line_Str;   /* fs */
+static char *Goto_Status_Line_Str;     /* ts */
+static int Num_Status_Line_Columns;    /* ws */
+static int Status_Line_Esc_Ok;	       /* es */
 
 /* static int Len_Curs_F_Str = 5; */
 
@@ -169,7 +219,7 @@ static int Scroll_r1 = 0, Scroll_r2 = 23;
 static int Cursor_r, Cursor_c;	       /* 0 based */
 
 /* current attributes --- initialized to impossible value */
-static SLtt_Char_Type Current_Fgbg = 0xFFFFFFFF;
+static SLtt_Char_Type Current_Fgbg = 0xFFFFFFFFU;
 
 static int Cursor_Set;		       /* 1 if cursor position known, 0
 					* if not.  -1 if only row is known
@@ -200,10 +250,10 @@ int SLtt_flush_output (void)
 	     if (errno == EAGAIN) 
 	       {
 #ifndef VMS
-		  struct timeval wait;
-		  wait.tv_sec = 0;
-		  wait.tv_usec = 100000;   /* 1/10 sec */
-		  (void) select(0, NULL, NULL, NULL, &wait);
+		  struct timeval tv;
+		  tv.tv_sec = 0;
+		  tv.tv_usec = 100000;   /* 1/10 sec */
+		  (void) select(0, NULL, NULL, NULL, &tv);
 #endif
 		  continue;
 	       }
@@ -212,10 +262,10 @@ int SLtt_flush_output (void)
 	     if (errno == EWOULDBLOCK)
 	       {
 #ifndef VMS
-		  struct timeval wait;
-		  wait.tv_sec = 0;
-		  wait.tv_usec = 100000;   /* 1/10 sec */
-		  (void) select(0, NULL, NULL, NULL, &wait);
+		  struct timeval tv;
+		  tv.tv_sec = 0;
+		  tv.tv_usec = 100000;   /* 1/10 sec */
+		  (void) select(0, NULL, NULL, NULL, &tv);
 #endif
 		  continue;
 	       }
@@ -233,14 +283,14 @@ int SLtt_flush_output (void)
 
 
 int SLtt_Baud_Rate;
-static void tt_write(char *str, int n)
+static void tt_write(char *str, unsigned int n)
 {
    static unsigned long last_time;
    static int total;
    unsigned long now;
-   int ndiff;
+   unsigned int ndiff;
    
-   if ((str == NULL) || (n <= 0)) return;
+   if ((str == NULL) || (n == 0)) return;
    total += n;
    
    while (1)
@@ -248,7 +298,7 @@ static void tt_write(char *str, int n)
 	ndiff = MAX_OUTPUT_BUFFER_SIZE - (int) (Output_Bufferp - Output_Buffer);
 	if (ndiff < n)
 	  {
-	     MEMCPY ((char *) Output_Bufferp, (char *) str, ndiff);
+	     SLMEMCPY ((char *) Output_Bufferp, (char *) str, ndiff);
 	     Output_Bufferp += ndiff;
 	     SLtt_flush_output ();
 	     n -= ndiff;
@@ -256,7 +306,7 @@ static void tt_write(char *str, int n)
 	  }
 	else 
 	  {
-	     MEMCPY ((char *) Output_Bufferp, str, n);
+	     SLMEMCPY ((char *) Output_Bufferp, str, n);
 	     Output_Bufferp += n;
 	     break;
 	  }
@@ -305,7 +355,7 @@ void SLtt_putchar(char ch)
 
 /* this is supposed to be fast--- also handles 
    termcap: %d, &i, %., %+, %r strings as well as terminfo stuff */
-static int tt_sprintf(char *buf, char *fmt, int x, int y)
+static unsigned int tt_sprintf(char *buf, char *fmt, int x, int y)
 {
    register unsigned char *f = (unsigned char *) fmt, *b, ch;
    int offset = 0, tinfo = 0;
@@ -343,19 +393,33 @@ static int tt_sprintf(char *buf, char *fmt, int x, int y)
 		    }
 	       }
 	     
-	     if (ch == 'p')
+	     switch (ch)
 	       {
+		case 'p':
 		  tinfo = 1;
 		  ch = *f++;
 		  if (ch == '1') stack[i++] = x; else stack[i++] = y;
-	       }
-	     else if (ch == '\'')   /* 'x' */
-	       {
+		  break;
+		  
+		case '\'':   /* 'x' */
 		  stack[i++] = *f++;
 		  f++;
-	       }
-	     else if ((ch == 'd') || (ch == '2') || (ch == '3'))
-	       {
+		  break;
+		  
+		case '{':	       /* literal constant, e.g. {30} */
+		  z = 0;
+		  while (((ch = *f) <= '9') && (ch >= '0'))
+		    {
+		       z = z * 10 + (ch - '0');
+		       f++;
+		    }
+		  stack[i++] = z;
+		  if (ch == '}') f++;
+		  break;
+		       
+		case 'd':
+		case '2':
+		case '3':
 		  z = stack[--i];
 		  z += offset;
 		  if (z >= 100)
@@ -379,13 +443,13 @@ static int tt_sprintf(char *buf, char *fmt, int x, int y)
 		  else if (ch == 2) *b++ = '0';
 		  
 		  *b++ = z + '0';
-	       }
-	     else if (ch == 'i') 
-	       {
+		  break;
+		  
+		case 'i':
 		  offset = 1;
-	       }
-	     else if (ch == '+')
-	       {
+		  break;
+		  
+		case '+':
 		  if (tinfo) 
 		    {
 		       z = stack[--i];
@@ -399,29 +463,33 @@ static int tt_sprintf(char *buf, char *fmt, int x, int y)
 		       if (ch == '\n') ch++;
 		       *b++ = ch;
 		    }
-	       }
-	     else if (ch == 'r')
-	       {
+		  break;
+		 
+		case 'r':
 		  stack[0] = x;
 		  stack[1] = y;
-	       }
-	     else if ((ch == '.') || (ch == 'c'))
-	       {
+		  break;
+		  
+		case '.':
+		case 'c':
 		  ch = (unsigned char) stack[--i];
 		  if (ch == '\n') ch++;
 		  *b++ = ch;
+		  break;
+		  
+		default:
+		  *b++ = ch;
 	       }
-	     else *b++ = ch;
 	  }
      }
    *b = 0;
-   return((int) (b - (unsigned char *) buf));
+   return((unsigned int) (b - (unsigned char *) buf));
 }
 
 static void tt_printf(char *fmt, int x, int y)
 {
    char buf[256];
-   int n;
+   unsigned int n;
    if (fmt == NULL) return;
    n = tt_sprintf(buf, fmt, x, y);
    tt_write(buf, n);
@@ -443,14 +511,10 @@ void SLtt_reset_scroll_region (void)
 
 int SLtt_set_cursor_visibility (int show)
 {
-   if (show) 
-     {
-	if (Cursor_Visible_Str != NULL) SLtt_write_string (Cursor_Visible_Str);
-	else return -1;
-	return 0;
-     }
-   if (Cursor_Invisible_Str != NULL) SLtt_write_string (Cursor_Invisible_Str);
-   else return -1;
+   if ((Cursor_Visible_Str == NULL) || (Cursor_Invisible_Str == NULL))
+     return -1;
+   
+   SLtt_write_string (show ? Cursor_Visible_Str : Cursor_Invisible_Str);
    return 0;
 }
 
@@ -543,14 +607,8 @@ void SLtt_delete_char (void)
 
 void SLtt_erase_line (void)
 {
-   char *s;
-   
-   Current_Fgbg = 0xFFFFFFFF;
    SLtt_write_string("\r");
    Cursor_Set = 1; Cursor_c = 0;
-   if (SLtt_Use_Ansi_Colors) s = "\033[0m";
-   else s = Norm_Vid_Str;
-   SLtt_write_string(s);
    SLtt_del_eol();
 }
 
@@ -558,7 +616,8 @@ void SLtt_delete_nlines (int n)
 {
    int r1, curs;
    char buf[132];
-   if (!n) return;
+   
+   if (n <= 0) return;
    SLtt_normal_video ();
    if (Del_N_Lines_Str != NULL) tt_printf(Del_N_Lines_Str,n, 0);
    else
@@ -568,8 +627,8 @@ void SLtt_delete_nlines (int n)
 	curs = Cursor_r;
 	SLtt_set_scroll_region(curs, Scroll_r2);
 	SLtt_goto_rc(Scroll_r2 - Scroll_r1, 0);
-	MEMSET(buf, '\n', n);
-	tt_write(buf, n);
+	SLMEMSET(buf, '\n', (unsigned int) n);
+	tt_write(buf, (unsigned int) n);
 	/* while (n--) tt_putchar('\n'); */
 	SLtt_set_scroll_region(r1, Scroll_r2);
 	SLtt_goto_rc(curs, 0);
@@ -621,7 +680,7 @@ void SLtt_beep (void)
 
 void SLtt_del_eol (void)
 {
-   if (Current_Fgbg != 0xFFFFFFFF) SLtt_normal_video ();
+   if (Current_Fgbg != 0xFFFFFFFFU) SLtt_normal_video ();
    SLtt_write_string(Del_Eol_Str);
 }
 
@@ -634,22 +693,22 @@ Color_Def_Type;
 
 static Color_Def_Type Color_Defs[16] =
 {
-   {"black",		0x00000000},
-   {"red",		0x00000100},
-   {"green",		0x00000200},
-   {"brown",		0x00000300},
-   {"blue",		0x00000400},
-   {"magenta",		0x00000500},
-   {"cyan",		0x00000600},
-   {"lightgray",	0x00000700},
-   {"gray",		0x00000800},
-   {"brightred",	0x00000900},
-   {"brightgreen",	0x00000A00},
-   {"yellow",		0x00000B00},
-   {"brightblue",	0x00000C00},
-   {"brightmagenta",	0x00000D00},
-   {"brightcyan",	0x00000E00},
-   {"white",		0x00000F00}
+   {"black",		SLSMG_COLOR_BLACK},
+   {"red",		SLSMG_COLOR_RED},
+   {"green",		SLSMG_COLOR_GREEN},
+   {"brown",		SLSMG_COLOR_BROWN},
+   {"blue",		SLSMG_COLOR_BLUE},
+   {"magenta",		SLSMG_COLOR_CYAN},
+   {"cyan",		SLSMG_COLOR_MAGENTA},
+   {"lightgray",	SLSMG_COLOR_LGRAY},
+   {"gray",		SLSMG_COLOR_GRAY},
+   {"brightred",	SLSMG_COLOR_BRIGHT_RED},
+   {"brightgreen",	SLSMG_COLOR_BRIGHT_GREEN},
+   {"yellow",		SLSMG_COLOR_BRIGHT_BROWN},
+   {"brightblue",	SLSMG_COLOR_BRIGHT_BLUE},
+   {"brightmagenta",	SLSMG_COLOR_BRIGHT_CYAN},
+   {"brightcyan",	SLSMG_COLOR_BRIGHT_MAGENTA},
+   {"white",		SLSMG_COLOR_BRIGHT_WHITE}
 };
 
 void SLtt_set_mono (int obj, char *what, SLtt_Char_Type mask)
@@ -682,6 +741,13 @@ void SLtt_set_color_object (int obj, SLtt_Char_Type attr)
    Ansi_Color_Map[obj].fgbg = attr;
 }
 
+SLtt_Char_Type SLtt_get_color_object (int obj)
+{
+   if ((obj < 0) || (obj >= JMAX_COLORS)) return 0;
+   return Ansi_Color_Map[obj].fgbg;
+}
+
+
 void SLtt_add_color_attribute (int obj, SLtt_Char_Type attr)
 {
    if ((obj < 0) || (obj >= JMAX_COLORS)) return;
@@ -693,46 +759,47 @@ void SLtt_add_color_attribute (int obj, SLtt_Char_Type attr)
 void SLtt_set_color (int obj, char *what, char *fg, char *bg)
 {
    int i;
-   SLtt_Char_Type f = 0xFFFFFFFF, g = 0xFFFFFFFF;
-   SLtt_Char_Type attr = 0;
+   SLtt_Char_Type f = 0xFFFFFFFFU, b = 0xFFFFFFFFU;
    
    if ((obj < 0) || (obj >= JMAX_COLORS))
      {
 	return;
      }
    
-   i = 0; while (i < 16)
+   for (i = 0; i < 16; i++)
      {
-	if (!strcmp(fg, Color_Defs[i].name))
-	  {
-	     f = Color_Defs[i].color;
-	     if (f & 0x800)
-	       {
-		  f &= 0x700;
-		  attr = SLTT_BOLD_MASK;
-	       }
-	     
-	     break;
-	  }
-	i++;
+	if (strcmp(fg, Color_Defs[i].name)) continue;
+	f = Color_Defs[i].color;
+	break;
      }
-   i = 0; while (i < 16)
+
+   for (i = 0; i < 16; i++)
      {
-	if (!strcmp(bg, Color_Defs[i].name))
-	  {
-	     g = Color_Defs[i].color;
-	     g &= 0x700;
-	     g = g << 8;
-	     break;
-	  }
-	i++;
+	if (strcmp(bg, Color_Defs[i].name)) continue;
+	b = Color_Defs[i].color;
+	break;
      }
    
-   if ((f == 0xFFFFFFFF) || (g == 0xFFFFFFFF) 
-       || ((f == g) && (attr == 0)))
+   if ((f == 0xFFFFFFFFU) || (b == 0xFFFFFFFFU))
      return;
 
-   SLtt_set_color_object (obj, f | g | attr);
+   SLtt_set_color_fgbg (obj, f, b);
+}
+
+void SLtt_set_color_fgbg (int obj, SLtt_Char_Type f, SLtt_Char_Type b)
+{
+   SLtt_Char_Type attr = 0;
+   
+   if (f & 0x8) attr = SLTT_BOLD_MASK;
+   f &= 0x7;
+   
+   if (b & 0x8) attr |= SLTT_BLINK_MASK;
+   b &= 0x7;
+       
+   if ((f == b) && (attr == 0))
+     return;
+   
+   SLtt_set_color_object (obj, (f << 8) | (b << 16) | attr);
 }
 
 void SLtt_set_color_esc (int obj, char *esc)
@@ -795,7 +862,7 @@ void SLtt_set_alt_char_set (int i)
 static void write_attributes (SLtt_Char_Type fgbg)
 {
    int bg0, fg0;
-   
+
    if (Worthless_Highlight) return;
    if (fgbg == Current_Fgbg) return;
    
@@ -855,7 +922,7 @@ void SLtt_reverse_video (int color)
 	     SLtt_write_string (Norm_Vid_Str);
 	  }
 	else SLtt_write_string (Rev_Vid_Str);
-	Current_Fgbg = 0xFFFFFFFF;
+	Current_Fgbg = 0xFFFFFFFFU;
 	return;
      }
    
@@ -916,7 +983,32 @@ void SLtt_wide_width (void)
 			    && COLOR_EQS((a), (b))))
 
 
-static void send_attr_str(unsigned short *s)
+/* The whole point of this routine is to prevent writing to the last column
+ * and last row on terminals with automatic margins.
+ */
+static void write_string_with_care (char *str)
+{
+   unsigned int len;
+   
+   if (str == NULL) return;
+   
+   len = strlen (str);
+   if (Automatic_Margins && (Cursor_r + 1 == SLtt_Screen_Rows))
+     {
+	if (len + Cursor_c >= SLtt_Screen_Cols)
+	  {
+	     /* For now, just do not write there.  Later, something more
+	      * sophisticated will be implemented.
+	      */
+	     if (SLtt_Screen_Cols > Cursor_c)
+	       len = SLtt_Screen_Cols - Cursor_c - 1;
+	     else len = 0;
+	  }
+     }
+   tt_write (str, len);
+}
+
+static void send_attr_str (unsigned short *s)
 {
    unsigned char out[256], ch, *p;
    register SLtt_Char_Type attr;
@@ -953,7 +1045,7 @@ static void send_attr_str(unsigned short *s)
 		       if (p != out)
 			 {
 			    *p = 0;
-			    SLtt_write_string((char *) out);
+			    write_string_with_care ((char *) out);
 			    Cursor_c += (int) (p - out);
 			    p = out;
 			 }
@@ -977,29 +1069,30 @@ static void send_attr_str(unsigned short *s)
 	*p++ = ch;
      }
    *p = 0;
-   if (p != out) SLtt_write_string((char *) out);
+   if (p != out) write_string_with_care ((char *) out);
    Cursor_c += (int) (p - out);
 }
 
-static void forward_cursor (int n, int row)
+static void forward_cursor (unsigned int n, int row)
 {
    char buf[30];
    
    
    if (n <= 4) 
      {
-	Cursor_c += n;
 	SLtt_normal_video ();
-	MEMSET (buf, ' ', n);
-	tt_write (buf, n);
+	SLMEMSET (buf, ' ', n);
+	buf[n] = 0;
+	write_string_with_care (buf);
+	Cursor_c += n;
      }
    else if (Curs_F_Str != NULL)
      {
 	Cursor_c += n;
-	n = tt_sprintf(buf, Curs_F_Str, n, 0);
+	n = tt_sprintf(buf, Curs_F_Str, (int) n, 0);
 	tt_write(buf, n);
      }
-   else SLtt_goto_rc (row, Cursor_c + n);
+   else SLtt_goto_rc (row, (int) (Cursor_c + n));
 }
 
 
@@ -1009,8 +1102,11 @@ void SLtt_smart_puts(unsigned short *neww, unsigned short *oldd, int len, int ro
 {
    register unsigned short *p, *q, *qmax, *pmax, *buf;
    unsigned short buffer[256];
-   int n_spaces;
+   unsigned int n_spaces;
    unsigned short *space_match, *last_buffered_match;
+#ifdef HP_GLITCH_CODE
+   int handle_hp_glitch = 0;
+#endif
    
    q = oldd; p = neww;
    qmax = oldd + len;
@@ -1027,6 +1123,24 @@ void SLtt_smart_puts(unsigned short *neww, unsigned short *oldd, int len, int ro
    /*position the cursor */
    SLtt_goto_rc (row, (int) (p - neww));
 
+#ifdef HP_GLITCH_CODE
+   if (Has_HP_Glitch)
+     {
+	unsigned short *qq = q;
+	while (qq < qmax)
+	  {
+	     if (*qq & 0xFF00)
+	       {
+		  SLtt_normal_video ();
+		  SLtt_del_eol ();
+		  qmax = q;
+		  handle_hp_glitch = 1;
+		  break;
+	       }
+	     qq++;
+	  }
+     }
+#endif
    /* Find where the last non-blank character on old/new screen is */
    
    while (qmax > q)
@@ -1049,9 +1163,18 @@ void SLtt_smart_puts(unsigned short *neww, unsigned short *oldd, int len, int ro
 	  }
      }
    
-   
    last_buffered_match = buf = buffer;		       /* buffer is empty */
-   
+
+#ifdef HP_GLITCH_CODE
+   if (handle_hp_glitch)
+     {
+	while (p < pmax)
+	  {
+	     *buf++ = *p++;
+	  }
+     }
+#endif
+
    /* loop using overwrite then skip algorithm until done */
    while (1)
      {
@@ -1074,7 +1197,7 @@ void SLtt_smart_puts(unsigned short *neww, unsigned short *oldd, int len, int ro
 		       p++;
 		       q++;
 		    }
-		  n_spaces = (int) (p - space_match);
+		  n_spaces = (unsigned int) (p - space_match);
 		  break;
 	       }
 	     if (CHAR_EQS(*q, *p)) break;
@@ -1110,10 +1233,11 @@ void SLtt_smart_puts(unsigned short *neww, unsigned short *oldd, int len, int ro
 	 */
 	if ((int) (buf - buffer) >= 5) 
 	  {
-	     forward_cursor (buf - buffer, row);
+	     forward_cursor ((unsigned int) (buf - buffer), row);
 	     last_buffered_match = buf = buffer;
 	  }
      }
+
    if (buf != buffer)
      {
 	if (q < qmax)
@@ -1121,7 +1245,7 @@ void SLtt_smart_puts(unsigned short *neww, unsigned short *oldd, int len, int ro
 	     if ((buf == last_buffered_match) 
 		 && ((int) (buf - buffer) >= 5))
 	       {
-		  forward_cursor (buf - buffer, row);
+		  forward_cursor ((unsigned int) (buf - buffer), row);
 	       }
 	     else
 	       {
@@ -1138,7 +1262,6 @@ void SLtt_smart_puts(unsigned short *neww, unsigned short *oldd, int len, int ro
 /* termcap stuff */
 
 #ifdef unix
-#define EXTERN extern
 
 #ifndef USE_TERMCAP
 static char *Tbuf;
@@ -1150,10 +1273,10 @@ static char *Tstr_Buf;
 #define TGETFLAG(x) SLtt_tigetflag((x), &Tbuf)
 #else
 
-EXTERN char *tgetstr(char *, char **);
-EXTERN int tgetent(char *, char *);
-EXTERN int tgetnum(char *);
-EXTERN int tgetflag(char *);
+extern char *tgetstr(char *, char **);
+extern int tgetent(char *, char *);
+extern int tgetnum(char *);
+extern int tgetflag(char *);
 static char Tstr_Buf[1024];
 static char Tbuf[4096];
 #define TGETNUM tgetnum
@@ -1167,6 +1290,10 @@ static char *my_tgetstr(char *what, char **p)
    what = tgetstr(what, p);
    if (what != NULL)
      {
+	/* Check for AIX brain-damage */
+	if (*what == '@')
+	  return NULL;
+	
 	/* lose pad info --- with today's technology, term is a loser if
 	   it is really needed */
 	while ((*what == '.') || 
@@ -1215,6 +1342,7 @@ void SLtt_get_terminfo (void)
 {
    char *term, *t, ch;
    int is_xterm;
+   int almost_vtxxx;
    
    SLtt_Use_Ansi_Colors = (NULL != getenv ("COLORTERM"));
 
@@ -1234,20 +1362,20 @@ void SLtt_get_terminfo (void)
    if (strcmp(t, "vt52") && (*t++ == 'v') && (*t++ == 't')
        && (ch = *t, (ch >= '1') && (ch <= '9'))) Vt100_Like = 1; 
 
+   is_xterm = !strncmp (term, "xterm", 5);
+   almost_vtxxx = (Vt100_Like
+		   || Linux_Console
+		   || is_xterm
+		   || !strcmp (term, "screen"));
+     
 #ifndef USE_TERMCAP
    if (NULL == (Tbuf = tgetent (term))) 
      {
 	char err_buf[512];
-	/* Special cases. */
-	if (Vt100_Like
-	    || Linux_Console
-	    || !strcmp (term, "screen"))
+	if (almost_vtxxx) /* Special cases. */
 	  {
 	     int vt102 = 1;
 	     if (!strcmp (term, "vt100")) vt102 = 0;
-	     SLtt_Screen_Cols = 80;
-	     SLtt_Screen_Rows = 24;
-	     if (Linux_Console) SLtt_Screen_Rows++;
    	     SLtt_set_term_vtxxx (&vt102);
 	     return;
 	  }
@@ -1285,10 +1413,20 @@ termcap entry.", term);
     */
    Term_Init_Str = SLtt_tgetstr ("ti");
    Term_Reset_Str = SLtt_tgetstr ("te");
-   
+
+   /* If I do this for vtxxx terminals, arrow keys start sending ESC O A, 
+    * which I do not want.  This is mainly for HP terminals.
+    */
+   if (almost_vtxxx == 0)
+     {
+	Keypad_Init_Str = SLtt_tgetstr ("ks");
+	Keypad_Reset_Str = SLtt_tgetstr ("ke");
+     }
+
    /* Make up for defective termcap/terminfo databases */
    if ((Vt100_Like && (term[2] != '1'))
        || Linux_Console
+       || is_xterm
        )
      {
 	if (Del_N_Lines_Str == NULL) Del_N_Lines_Str = "\033[%dM";
@@ -1297,36 +1435,18 @@ termcap entry.", term);
    
    Scroll_R_Str = SLtt_tgetstr("cs");
    
-   if ((SLtt_Screen_Rows == 0) || (SLtt_Screen_Cols == 0))
-     {
-	char *envptr;
-	if ((SLtt_Screen_Cols = TGETNUM("co")) <= 0) SLtt_Screen_Cols = 80;
-	if ((SLtt_Screen_Rows = TGETNUM("li")) <= 0) SLtt_Screen_Rows = 24;
-	envptr = getenv ("LINES");
-	if (envptr != NULL)
-	  {
-	     if ((SLtt_Screen_Rows = atoi (envptr)) <= 0) 
-	       SLtt_Screen_Rows = 24;
-	  }
-	
-	envptr = getenv ("COLUMNS");
-	if (envptr != NULL)
-	  {
-	     if ((SLtt_Screen_Cols = atoi (envptr)) <= 0) 
-	       SLtt_Screen_Cols = 80;
-	  }
-     }
-   
-   is_xterm = !strncmp (term, "xterm", 5);
-   
+   SLtt_get_screen_size ();
+
    if ((Scroll_R_Str == NULL) 
        || (((NULL == Del_N_Lines_Str) || (NULL == Add_N_Lines_Str))
 	   && (NULL == Rev_Scroll_Str)))
      {
-	if (is_xterm)
+	if (is_xterm
+	    || Linux_Console
+	    )
 	  {
 	     /* Defective termcap mode!!!! */
-	     SLtt_set_term_vtxxx (&is_xterm);
+	     SLtt_set_term_vtxxx (NULL);
 	  }
 	else SLtt_Term_Cannot_Scroll = 1;
      }
@@ -1337,45 +1457,47 @@ termcap entry.", term);
    if (Rev_Vid_Str == NULL) Rev_Vid_Str = SLtt_tgetstr("so");
    
    Bold_Vid_Str = SLtt_tgetstr("md");
-   Blink_Vid_Str = SLtt_tgetstr("mb");
+
+   /* Although xterm cannot blink, it does display the blinking characters
+    * as bold ones.  Some Rxvt will display the background as high intensity.
+    */
+   if ((NULL == (Blink_Vid_Str = SLtt_tgetstr("mb")))
+       && is_xterm)
+     Blink_Vid_Str = "\033[5m";
+   
    UnderLine_Vid_Str = SLtt_tgetstr("us");
    
    Start_Alt_Chars_Str = SLtt_tgetstr ("as");   /* smacs */
    End_Alt_Chars_Str = SLtt_tgetstr ("ae");   /* rmacs */
    Enable_Alt_Char_Set = SLtt_tgetstr ("eA");   /* enacs */
    SLtt_Graphics_Char_Pairs = SLtt_tgetstr ("ac");
+
+#ifndef NCURSES_BRAIN_DAMAGE_CONTROL
+# define NCURSES_BRAIN_DAMAGE_CONTROL 0
+#endif
    
-   /* I have had it with defective console/linux terminfo entries provided 
-    * by ncurses.  They are not reliable.  I cannot track every ncurses 
-    * release and fudge my stuff to work with defective stuff.  For that
-    * reason, I am hardcoding some things.
-    */
+#if NCURSES_BRAIN_DAMAGE_CONTROL
    if (Linux_Console)
      {
-#if 0
+# if 0
 	char *lgcp = "l\332m\300k\277j\331u\264t\303v\301w\302q\304x\263n\053o\176s\137`\004a\260f\370g\361~\011,\020+\021.\031-\030h\261i\0250\333";
 
-	if ((SLtt_Graphics_Char_Pairs == NULL)
-	    || (0 == strncmp (SLtt_Graphics_Char_Pairs, lgcp, 8)))
-	  {
-	     Start_Alt_Chars_Str = "\033(B\033)U\016";
-	     End_Alt_Chars_Str = "\033(B\033)0\017";
-	     Enable_Alt_Char_Set = NULL;
-	  }
-#else
+	SLtt_Graphics_Char_Pairs = lgcp;
+	Start_Alt_Chars_Str = "\033(B\033)U\016";
+	End_Alt_Chars_Str = "\033(B\033)0\017";
+	Enable_Alt_Char_Set = NULL;
+# else
 	char *lgcp = "`\004a\261f\370g\361h\260j\331k\277l\332m\300n\305o\302q\304r\362s_t\303u\264v\301w\302x\263y\371z\372{\373|\374}\375~";
 	
-	if ((SLtt_Graphics_Char_Pairs == NULL)
-	    || (0 == strncmp (SLtt_Graphics_Char_Pairs, lgcp, 16)))
-	  {
-	     SLtt_Graphics_Char_Pairs = lgcp;
-	     Start_Alt_Chars_Str = "\033[11m";
-	     End_Alt_Chars_Str = "\033[10m";
-	     Enable_Alt_Char_Set = NULL;
-	  }
-#endif
+	SLtt_Graphics_Char_Pairs = lgcp;
+	Start_Alt_Chars_Str = "\033[11m";
+	End_Alt_Chars_Str = "\033[10m";
+	Enable_Alt_Char_Set = NULL;
+# endif
      }
-   else if (NULL == SLtt_Graphics_Char_Pairs)
+#endif
+   
+   if (NULL == SLtt_Graphics_Char_Pairs)
      {
 	/* make up for defective termcap/terminfo */
 	if (Vt100_Like)
@@ -1386,7 +1508,8 @@ termcap entry.", term);
 	  }
      }
    
-   if (is_xterm)
+    /* aixterm added by willi */
+   if (is_xterm || !strncmp (term, "aixterm", 7))
      {
 	Start_Alt_Chars_Str = "\016";
 	End_Alt_Chars_Str = "\017";
@@ -1401,7 +1524,18 @@ termcap entry.", term);
      }
    else SLtt_Has_Alt_Charset = 1;
    
-       
+
+   /* status line capabilities */
+   if ((SLtt_Has_Status_Line == -1)
+       && (0 != (SLtt_Has_Status_Line = TGETFLAG ("hs"))))
+     {
+	Disable_Status_line_Str = SLtt_tgetstr ("ds");
+	Return_From_Status_Line_Str = SLtt_tgetstr ("fs");
+	Goto_Status_Line_Str = SLtt_tgetstr ("ts");
+	Status_Line_Esc_Ok = TGETFLAG("es");
+	Num_Status_Line_Columns = TGETNUM("ws");
+	if (Num_Status_Line_Columns < 0) Num_Status_Line_Columns = 0;
+     }
    
    if (NULL == (Norm_Vid_Str = SLtt_tgetstr("me"))) 
      {
@@ -1409,7 +1543,7 @@ termcap entry.", term);
      }
    
    Cursor_Invisible_Str = SLtt_tgetstr("vi");
-   Cursor_Visible_Str = SLtt_tgetstr("vs");
+   Cursor_Visible_Str = SLtt_tgetstr("ve");
    
    Curs_F_Str = SLtt_tgetstr("RI");
    
@@ -1423,12 +1557,36 @@ termcap entry.", term);
    
    Automatic_Margins = TGETFLAG ("am");
    /* No_Move_In_Standout = !TGETFLAG ("ms"); */
+#ifdef HP_GLITCH_CODE
+   Has_HP_Glitch = TGETFLAG ("xs");
+#else
    Worthless_Highlight = TGETFLAG ("xs");
-   
+#endif
+
    if (Worthless_Highlight == 0)
-     {
+     {				       /* Magic cookie glitch */
 	Worthless_Highlight = (TGETNUM ("sg") > 0);
      }
+
+   if (Worthless_Highlight)
+     SLtt_Has_Alt_Charset = 0;
+   
+   /* Check for color information in the termcap.  A program should not
+    * rely on this information being accurate.
+    */
+   if (SLtt_Use_Ansi_Colors == 0)
+     {
+	Reset_Color_String = SLtt_tgetstr ("op");
+	
+	SLtt_Use_Ansi_Colors = ((NULL != Reset_Color_String)
+				|| (NULL != SLtt_tgetstr ("Sf"))
+				|| (NULL != SLtt_tgetstr ("Sb"))
+				|| (NULL != SLtt_tgetstr ("AF"))
+				|| (NULL != SLtt_tgetstr ("AB"))
+				|| (-1 != SLtt_tgetnum ("Co"))
+				|| (-1 != SLtt_tgetnum ("pa")));
+	
+     }
 }
 
 #endif
@@ -1453,6 +1611,8 @@ void SLtt_get_terminfo ()
    SLtt_Has_Alt_Charset = 1;
    SLtt_Graphics_Char_Pairs = "aaffgghhjjkkllmmnnooqqssttuuvvwwxx";
    Enable_Alt_Char_Set = "\033(B\033)0"; 
+   SLtt_Use_Ansi_Colors = (NULL != getenv ("COLORTERM"));
+   SLtt_get_screen_size ();
 }
 #endif
 
@@ -1475,7 +1635,7 @@ void SLtt_set_term_vtxxx(int *vt100)
    Curs_F_Str = "\033[%dC";
    /* Len_Curs_F_Str = 5; */
    Curs_Pos_Str = "\033[%i%d;%dH";
-   if (*vt100 == 0)
+   if ((vt100 == NULL) || (*vt100 == 0))
      {
 	Ins_Mode_Str = "\033[4h";
 	Eins_Mode_Str = "\033[4l";
@@ -1499,6 +1659,7 @@ void SLtt_init_video (void)
    /*   send_string_to_term("\033[?6h"); */
    /* relative origin mode */
    SLtt_write_string (Term_Init_Str);
+   SLtt_write_string (Keypad_Init_Str);
    SLtt_reset_scroll_region();
    SLtt_end_insert();
    SLtt_write_string (Enable_Alt_Char_Set);
@@ -1512,10 +1673,18 @@ void SLtt_reset_video (void)
    Cursor_Set = 0;
    SLtt_normal_video ();	       /* MSKermit requires this  */
    SLtt_write_string(Norm_Vid_Str);
-   Current_Fgbg = 0xFFFFFFFF;
+   Current_Fgbg = 0xFFFFFFFFU;
    SLtt_set_alt_char_set (0);
-   SLtt_write_string ("\n");
-   SLtt_del_eol ();
+   if (SLtt_Use_Ansi_Colors) 
+     {
+	if (Reset_Color_String == NULL)
+	  {
+	     Reset_Color_String = "\033[0m\033[m";
+	  }
+	SLtt_write_string (Reset_Color_String);
+     }
+   SLtt_erase_line ();
+   SLtt_write_string (Keypad_Reset_Str);
    SLtt_write_string (Term_Reset_Str);
    SLtt_flush_output ();
    Video_Initialized = 0;
@@ -1536,7 +1705,7 @@ int SLtt_set_mouse_mode (int mode, int force)
 	if (strncmp ("xterm", term, 5))
 	  return -1;
      }
-   
+
    if (mode)
      SLtt_write_string ("\033[?9h");
    else
@@ -1544,3 +1713,95 @@ int SLtt_set_mouse_mode (int mode, int force)
    
    return 0;
 }
+
+
+void SLtt_disable_status_line (void)
+{
+   if (SLtt_Has_Status_Line > 0)
+     SLtt_write_string (Disable_Status_line_Str);
+}
+
+int SLtt_write_to_status_line (char *s, int col)
+{
+   if ((SLtt_Has_Status_Line <= 0)
+       || (Goto_Status_Line_Str == NULL)
+       || (Return_From_Status_Line_Str == NULL))
+     return -1;
+   
+   tt_printf (Goto_Status_Line_Str, col, 0);
+   SLtt_write_string (s);
+   SLtt_write_string (Return_From_Status_Line_Str);
+   return 0;
+}
+
+   
+void SLtt_get_screen_size (void)
+{   
+#ifdef VMS
+   int status, code;
+   unsigned short chan;
+   $DESCRIPTOR(dev_dsc, "SYS$INPUT:");
+#endif
+#ifdef __os2__
+   VIOMODEINFO vioModeInfo;
+#endif
+   int r = 0, c = 0;
+   
+#ifdef TIOCGWINSZ
+   struct winsize wind_struct;
+   
+   do
+     {
+	if ((ioctl(1,TIOCGWINSZ,&wind_struct) == 0)
+	    || (ioctl(0, TIOCGWINSZ, &wind_struct) == 0)
+	    || (ioctl(2, TIOCGWINSZ, &wind_struct) == 0))
+	  {
+	     c = (int) wind_struct.ws_col;
+	     r = (int) wind_struct.ws_row;
+	     break;
+	  }
+     }
+   while (errno == EINTR);
+   
+#endif
+
+#ifdef VMS
+   status = sys$assign(&dev_dsc,&chan,0,0,0);
+   if (status & 1)
+     {
+	code = DVI$_DEVBUFSIZ;
+	status = lib$getdvi(&code, &chan,0, &c, 0,0);
+	if (!(status & 1))
+	  c = 80;
+	code = DVI$_TT_PAGE;
+	status = lib$getdvi(&code, &chan,0, &r, 0,0);
+	if (!(status & 1))
+	  r = 24;
+	sys$dassgn(chan);
+     }
+#endif
+
+#ifdef __os2__
+   vioModeInfo.cb = sizeof(vioModeInfo);
+   VioGetMode (&vioModeInfo, 0);
+   c = vioModeInfo.col;
+   r = vioModeInfo.row;   
+#endif
+   
+   if (r <= 0)
+     {
+	char *s = getenv ("LINES");
+	if (s != NULL) r = atoi (s);
+     }
+   
+   if (c <= 0)
+     {
+	char *s = getenv ("COLUMNS");
+	if (s != NULL) c = atoi (s);
+     }
+   
+   if ((r <= 0) || (r > 200)) r = 24;
+   if ((c <= 0) || (c > 250)) c = 80;
+   SLtt_Screen_Rows = r;
+   SLtt_Screen_Cols = c;
+}
