@@ -11,12 +11,13 @@
  */
 
 
+#include "config.h"
+
 #include <stdio.h>
 #include <string.h>
 
-#include "config.h"
 
-#ifndef NO_STDLIB_H
+#ifdef HAVE_STDLIB_H
 # include <stdlib.h>
 #endif
 
@@ -48,7 +49,7 @@ static SLang_Key_Type *malloc_key(unsigned char *str)
 	SLang_Error = SL_MALLOC_ERROR;
 	return NULL;
      }
-   MEMCPY((char *) neew->str, (char *) str, (int) *str);
+   SLMEMCPY((char *) neew->str, (char *) str, (unsigned int) *str);
    return(neew);
 }
 
@@ -72,57 +73,71 @@ static SLKeyMap_List_Type *add_keymap(char *name, SLang_Key_Type *map)
    return NULL;
 }
 
-VOID_STAR SLang_find_key_function(char *name, SLKeyMap_List_Type *keymap)
+#ifdef SLKEYMAP_OBSOLETE
+VOID_STAR 
+#else
+FVOID_STAR
+#endif
+SLang_find_key_function(char *name, SLKeyMap_List_Type *keymap)
 {
    SLKeymap_Function_Type *fp = keymap -> functions;
    
-   while((fp != NULL) && (fp->name != NULL))
+   while ((fp != NULL) && (fp->name != NULL))
      {
-	if ((*fp->name == *name) && !strcmp(fp->name, name)) return((VOID_STAR) fp->f);
+	if ((*fp->name == *name) && !strcmp(fp->name, name)) 
+	  {
+#ifdef SLKEYMAP_OBSOLETE
+	     return (VOID_STAR) fp->f;
+#else
+	     return (FVOID_STAR) fp->f;
+#endif
+	  }
 	fp++;
      }
    return(NULL);
 }
 
-/* convert things like "^A" to 1 etc... The 0th char is the strlen */
+/* convert things like "^A" to 1 etc... The 0th char is the strlen INCLUDING
+ * the length character itself.
+ */
 char *SLang_process_keystring(char *s)
 {
    static char str[30];
    unsigned char ch;
    int i;
 
-    i = 1;
-    while (*s != 0)
-      {
-          ch = (unsigned char) *s++;
-          if (*s) ch = UPPER_CASE_KEY(ch);
-          if (ch == '^')
-            {
-                ch = *s++;
-                if (ch == 0)
-                  {
-                      str[i++] = '^';
-                      break;
-                  }
-	       if (ch == '?') ch = 127; else ch = ch - 'A' + 1;
-            }
-
-          str[i++] = ch;
-      }
-    str[0] = i;
-    return(str);
+   i = 1;
+   while (*s != 0)
+     {
+	ch = (unsigned char) *s++;
+	if (*s) ch = UPPER_CASE_KEY(ch);
+	if (ch == '^')
+	  {
+	     ch = *s++;
+	     if (ch == 0)
+	       {
+		  str[i++] = '^';
+		  break;
+	       }
+	     if (ch == '?') ch = 127; else ch = ch - 'A' + 1;
+	  }
+	
+	str[i++] = ch;
+     }
+   str[0] = i;
+   return(str);
 }
 
 
 static char *Define_Key_Error = "Inconsistency in define key.";
 
-/* returns -2 if inconsistent, -1 if malloc error, 0 upon success */
-int SLang_define_key1(char *s, VOID_STAR f, unsigned char type, SLKeyMap_List_Type *kml)
+static int find_the_key (char *s, SLKeyMap_List_Type *kml, SLang_Key_Type **keyp)
 {
    int cmp, i, m, n, len;
    SLang_Key_Type *key, *last, *neew;
    unsigned char *str;
 
+   *keyp = NULL;
    str = (unsigned char *) SLang_process_keystring(s);
    if (1 == (n = str[0])) return 0;
 
@@ -139,11 +154,17 @@ int SLang_define_key1(char *s, VOID_STAR f, unsigned char type, SLKeyMap_List_Ty
 	
 	/* copy keymap uses the pointers so do not free these if copied */
 	if ((SLKeyMap_List[1].keymap == NULL)
-	    && (key->type == SLKEY_F_INTERPRET)) SLFREE(key->f);
-	
-	key->f = f;
-	key->type = type;
-	MEMCPY((char *) key->str, (char *) str, 2);
+	    && (key->type == SLKEY_F_INTERPRET))
+	  {
+#ifdef SLKEYMAP_OBSOLETE
+	     SLFREE(key->f);
+#else
+	     SLFREE (key->f.s);
+#endif
+	  }
+
+	SLMEMCPY((char *) key->str, (char *) str, 2);
+	*keyp = key;
 	return 0;
      }
 
@@ -158,59 +179,107 @@ int SLang_define_key1(char *s, VOID_STAR f, unsigned char type, SLKeyMap_List_Ty
 	     m = key->str[0];
 	     len = m;
 	     if (m > n) len = n;
-	     cmp = MEMCMP((char *) (str + 1), (char *)(key->str + 1), len - 1);
+	     cmp = SLMEMCMP((char *) (str + 1), (char *)(key->str + 1), len - 1);
 	     if ((m != n) && !cmp)
 	       {
 		  SLang_doerror (Define_Key_Error); 
 		  return -2;
 	       }
-	     if ((m == n) && (!cmp)) cmp = MEMCMP((char *) str, (char *) key->str, n);
+	     if ((m == n) && (!cmp)) cmp = SLMEMCMP((char *) str, (char *) key->str, n);
 	  }
 
 	if (cmp == 0)
 	  {
 	       /* copy keymap uses the pointers so do not free these if copied */
-	       if ((SLKeyMap_List[1].keymap == NULL)
-		   && (key->type == SLKEY_F_INTERPRET)) SLFREE(key->f);
-	       key->f = f;
-	       key->type = type;
-	       return 0;
-            }
-          else if (cmp < 0)
-            {
-	       if (NULL == (neew = malloc_key(str))) return -1;
-
-	       neew -> next = key;
-	       last -> next = neew;
-
-	       neew->f = f;
-	       neew->type = type;
-	       return 0;
-            } /* else cmp > 0 */
+	     if ((SLKeyMap_List[1].keymap == NULL)
+		 && (key->type == SLKEY_F_INTERPRET))
+	       {
+#ifdef SLKEYMAP_OBSOLETE	
+		  SLFREE(key->f);
+#else
+		  SLFREE(key->f.s);
+#endif
+	       }
+	     
+	     *keyp = key;
+	     return 0;
+	  }
+	else if (cmp < 0)
+	  {
+	     if (NULL == (neew = malloc_key(str))) return -1;
+	     
+	     neew -> next = key;
+	     last -> next = neew;
+	     
+	     *keyp = neew;
+	     return 0;
+	  } /* else cmp > 0 */
       }
 }
 
+/* returns -2 if inconsistent, -1 if malloc error, 0 upon success */
+#ifdef SLKEYMAP_OBSOLETE
+int SLang_define_key1 (char *s, VOID_STAR f, unsigned int type, SLKeyMap_List_Type *kml)
+#else
+int SLkm_define_key (char *s, FVOID_STAR f, SLKeyMap_List_Type *kml)
+#endif
+{
+   SLang_Key_Type *key;
+#ifndef SLKEYMAP_OBSOLETE
+   unsigned int type = SLKEY_F_INTRINSIC;
+#endif
+   int ret;
+   
+   ret = find_the_key (s, kml, &key);
+   if ((ret != 0) || (key == NULL))
+     return ret;
+
+   key->type = type;
+#ifdef SLKEYMAP_OBSOLETE
+   key->f = f;
+#else
+   key->f.f = f;
+#endif
+   return 0;
+}
+
 int SLang_define_key(char *s, char *funct, SLKeyMap_List_Type *kml)
 {
+   SLang_Key_Type *key;
+#ifdef SLKEYMAP_OBSOLETE
    VOID_STAR f;
-   unsigned char type;
+#else
+   FVOID_STAR f;
+#endif
+   int ret;
+   
+   ret = find_the_key (s, kml, &key);
+   if ((ret != 0) || (key == NULL))
+     return ret;
 
    f = SLang_find_key_function(funct, kml);
+
    if (f == NULL)                      /* assume interpreted */
      {
-	if (NULL == (f = (VOID_STAR) SLMALLOC(strlen(funct) + 1)))
-	  {
-	     SLang_Error = SL_MALLOC_ERROR;
-	     SLang_doerror ("Malloc error in define key.");
-	     return -1;
-	  }
-
-	strcpy((char *) f, funct);
-	type = SLKEY_F_INTERPRET;
+	char *str = SLmake_string (funct);
+	if (str == NULL) return -1;
+	key->type = SLKEY_F_INTERPRET;
+#ifdef SLKEYMAP_OBSOLETE
+	key->f = (VOID_STAR) str;
+#else
+	key->f.s = str;
+#endif
      }
-   else type = SLKEY_F_INTRINSIC;
-
-   return SLang_define_key1(s, f, type, kml);
+   else 
+     {
+	key->type = SLKEY_F_INTRINSIC;
+#ifdef SLKEYMAP_OBSOLETE
+	key->f = f;
+#else
+	key->f.f = f;
+#endif
+     }
+   return 0;
 }
 
 
@@ -233,7 +302,14 @@ SLang_Key_Type *SLang_do_key(SLKeyMap_List_Type *kml, int (*getkey)(void))
    if (key->next == NULL) 
      {
 	/* check its uppercase counterpart if null and return it */
-	if ((key->f == NULL) && (i >= 'a') && (i <= 'z')) key = (SLang_Key_Type *) &((kml->keymap)[i - 32]);
+	if (
+#ifdef SLKEYMAP_OBSOLETE
+	    (key->f == NULL)
+#else
+	    (key->f.f == NULL)
+#endif
+	    && (i >= 'a') && (i <= 'z')) 
+	  key = (SLang_Key_Type *) &((kml->keymap)[i - 32]);
 	SLang_Key_TimeOut_Flag = 0;
 	return(key);
      }
@@ -332,10 +408,17 @@ void SLang_undefine_key(char *s, SLKeyMap_List_Type *kml)
    while (key != NULL)
      {
 	next = key->next;
-	if (!MEMCMP((char *)(key->str + 1), (char *) str, n))
+	if (!SLMEMCMP((char *)(key->str + 1), (char *) str, n))
 	  {
 	     if ((SLKeyMap_List[1].keymap == NULL)
-		 && (key->type == SLKEY_F_INTERPRET)) SLFREE(key->f);
+		 && (key->type == SLKEY_F_INTERPRET))
+	       {
+#ifdef SLKEYMAP_OBSOLETE
+		  SLFREE(key->f);
+#else
+		  SLFREE (key->f.s);
+#endif
+	       }
 	     SLFREE(key);
 	     last->next = next;
 	  }
@@ -346,7 +429,11 @@ void SLang_undefine_key(char *s, SLKeyMap_List_Type *kml)
    if (n == 1)
      {
 	*key_root->str = 0;
+#ifdef SLKEYMAP_OBSOLETE
 	key_root->f = NULL;
+#else
+	key_root->f.f = NULL;
+#endif
 	key_root->type = 0;
      }
 }
@@ -395,16 +482,30 @@ static SLang_Key_Type *copy_keymap(SLKeyMap_List_Type *kml)
      {
 	old = &(km[i]);
 	neew = &(new_root[i]);
+#ifdef SLKEYMAP_OBSOLETE
 	neew->f = old->f;
+#else
+	if (old->type == SLKEY_F_INTERPRET)
+	  neew->f.s = old->f.s;
+	else
+	  neew->f.f = old->f.f;
+#endif
 	neew->type = old->type;
-	MEMCPY((char *) neew->str, (char *) old->str, (int) *old->str);
+	SLMEMCPY((char *) neew->str, (char *) old->str, (unsigned int) *old->str);
 
 	old = old->next;
 	while (old != NULL)
 	  {
 	     neew->next = malloc_key((unsigned char *) old->str);
 	     neew = neew->next;
+#ifdef SLKEYMAP_OBSOLETE
 	     neew->f = old->f;
+#else
+	     if (old->type == SLKEY_F_INTERPRET)
+	       neew->f.s = old->f.s;
+	     else
+	       neew->f.f = old->f.f;
+#endif
 	     neew->type = old->type;
 	     old = old->next;
 	  }
