@@ -7,10 +7,11 @@
  */
 
 
+#include "config.h"
+
 #include <stdio.h>
 #include <string.h>
 
-#include "config.h"
 
 #include "slang.h"
 
@@ -27,6 +28,9 @@
 #define ANY_DIGIT 0x9		       /* \d */
 #define BOW	0xA		       /* \< */
 #define EOW	0xB		       /* \> */
+#if 0
+#define NOT_LITERAL		0xC	       /* \~ */
+#endif
 #define STAR 0x80		       /* * */
 #define LEAST_ONCE 0x40		       /* + */
 #define MAYBE_ONCE 0x20		       /* ? */
@@ -56,7 +60,7 @@ static unsigned char *do_nth_match(int n, unsigned char *str, unsigned char *est
    if (n > (int) (estr - str)) return (NULL);
    
    /* This needs fixed for case sensitive match */
-   if (0 != strncmp((char *) str, (char *) bpos, n)) return (NULL);
+   if (0 != strncmp((char *) str, (char *) bpos, (unsigned int) n)) return (NULL);
    str += n;
    return (str);
 }
@@ -99,10 +103,56 @@ static unsigned char *regexp_looking_at(register unsigned char *str, unsigned ch
 	     This_Reg->beg_matches[Next_Pos + 1] = -1;
 	     break;
 	   case CPAREN: 
-	     This_Reg->end_matches[Next_Pos] = (int) (str - This_Str) - This_Reg->beg_matches[Next_Pos];
+	     This_Reg->end_matches[Next_Pos] = (unsigned int) (str - This_Str) - This_Reg->beg_matches[Next_Pos];
 	     Next_Pos++;
 	     break;
+#ifdef NOT_LITERAL
+	   case NOT_LITERAL:
+	     if ((str >= estr) || (*buf == UPPERCASE(*str))) return (NULL);
+	     str++; buf++;
+	     break;
 	     
+	   case MAYBE_ONCE | NOT_LITERAL:
+	     save_str = str;
+	     if ((str < estr) && (*buf != UPPERCASE(*str))) str++;
+	     buf++;
+	     goto match_rest;
+
+	   case NOT_LITERAL | LEAST_ONCE:   /* match at least once */
+	     if ((str >= estr) || (UPPERCASE(*str) == UPPERCASE(*buf))) return (NULL);
+	     str++;
+	     /* drop */
+	   case STAR | NOT_LITERAL:
+	     save_str = str;  p1 = *buf;
+	     while ((str < estr) && (UPPERCASE(*str) != p1)) str++;
+	     buf++;
+	     goto match_rest;
+	     
+	     /* this type consists of the expression + two bytes that
+	        determine number of matches to perform */
+	   case MANY | NOT_LITERAL:
+	     p1 = *buf; buf++;
+	     n = n0 = (int) (unsigned char) *buf++;
+	     /* minimum number to match--- could be 0 */
+	     n1 = (int) (unsigned char) *buf++;
+	     /* maximum number to match */
+
+	     while (n && (str < estr) && (p1 != *str)) 
+	       {
+		  n--;
+		  str++;
+	       }
+	     if (n) return (NULL);
+	     
+	     save_str = str;
+	     n = n1 - n0;
+	     while (n && (str < estr) && (p1 != *str)) 
+	       {
+		  n--;
+		  str++;
+	       }
+	     goto match_rest;
+#endif				       /* NOT_LITERAL */
 	   case LITERAL: 
 	     if ((str >= estr) || (*buf != UPPERCASE(*str))) return (NULL);
 	     str++; buf++;
@@ -309,7 +359,8 @@ static unsigned char *regexp_looking_at(register unsigned char *str, unsigned ch
 }
 
 
-unsigned char *SLang_regexp_match(unsigned char *str, int len, SLRegexp_Type *reg)
+unsigned char *SLang_regexp_match(unsigned char *str, 
+				  unsigned int len, SLRegexp_Type *reg)
 {
    register unsigned char c = 0, *estr = str + len;
    int cs = reg->case_sensitive, lit = 0;
@@ -325,7 +376,7 @@ unsigned char *SLang_regexp_match(unsigned char *str, int len, SLRegexp_Type *re
      {
 	if (NULL == (epos = regexp_looking_at(str, estr, buf + 1, cs))) str = NULL;
 	reg->beg_matches[0] = (int) (str - This_Str);
-	reg->end_matches[0] = (int) (epos - str);
+	reg->end_matches[0] = (unsigned int) (epos - str);
 	return(str);
      }
 
@@ -362,7 +413,7 @@ unsigned char *SLang_regexp_match(unsigned char *str, int len, SLRegexp_Type *re
 	if (NULL != (epos = regexp_looking_at(str, estr, buf, cs)))
 	  {
 	     reg->beg_matches[0] = (int) (str - This_Str);
-	     reg->end_matches[0] = (int) (epos - str);
+	     reg->end_matches[0] = (unsigned int) (epos - str);
 	     return (str);
 	  }
 	str++;
@@ -406,7 +457,7 @@ int SLang_regexp_compile (SLRegexp_Type *reg)
    
    reg->beg_matches[0] = reg->end_matches[0] = 0;
    buf = reg->buf;
-   ebuf = reg->buf + reg->buf_len - 2; /* make some room */
+   ebuf = (reg->buf + reg->buf_len) - 2; /* make some room */
    pat = reg->pat;
    cs = reg->case_sensitive;
 
@@ -423,7 +474,8 @@ int SLang_regexp_compile (SLRegexp_Type *reg)
    
    i = 1; while (i < 10) 
      {
-	reg->beg_matches[i] = reg->end_matches[i] = -1;
+	reg->beg_matches[i] = -1;
+	reg->end_matches[i] = 0;
 	i++;
      }
 
@@ -442,7 +494,7 @@ int SLang_regexp_compile (SLRegexp_Type *reg)
      {
 	if (buf >= ebuf - 3)
 	  {
-	     SLang_doerror ("Pattern large to be compiled.");
+	     SLang_doerror ("Pattern too large to be compiled.");
 	     ERROR;
 	  }
 	
@@ -471,7 +523,16 @@ int SLang_regexp_compile (SLRegexp_Type *reg)
 		  last = buf;
 		  *buf++ = NTH_MATCH; *buf++ = c;
 		  break;
-
+#ifdef NOT_LITERAL
+		case '~':	       /* slang extension */
+		  if ((c = *pat) == 0) ERROR;
+		  pat++;
+		  last = buf;
+		  *buf++ = NOT_LITERAL;
+		  *buf++ = c;
+		  min_length++;
+		  break;
+#endif
 		case 'd':	       /* slang extension */
 		  last = buf;
 		  *buf++ = ANY_DIGIT;
@@ -686,7 +747,7 @@ int SLang_regexp_compile (SLRegexp_Type *reg)
    
    reg->osearch = ordinary_search;
    reg->must_match_str[15] = 0;
-   reg->min_length = (min_length > 0) ? min_length : 0;
+   reg->min_length = (min_length > 0) ? (unsigned int) min_length : 0;
    if (ordinary_search)
      {
 	strncpy((char *) reg->must_match_str, (char *) reg->pat, 15);
